#include <math.h>
#include "unifloat.h"
#include "trig.h"
#include "exp.h"
#include "debug.h"
#include "bessel.h"
#include "gamma.h"


/* Functions to provide:
ceil	округление до ближайшего большего целого числа
cosh
fabs	абсолютная величина (числа с плавающей точкой)
floor	округление до ближайшего меньшего целого числа
fmod	вычисление остатка от деления нацело для чисел с плавающей точкой
frexp	разбивает число с плавающей точкой на мантиссу и показатель степени.
ldexp	умножение числа с плавающей точкой на целую степень двух
log10	логарифм по основанию 10
odf(x,p)	извлекает целую и дробную части (с учетом знака) из числа с плавающей точкой
pow(x,y)	результат возведения x в степень y, x
sinh	гиперболический синус
tanh	гиперболический тангенс
acosh	гиперболический арккосинус
asinh	гиперболический арксинус
atanh	гиперболический арктангенс
cbrt	кубический корень
copysign(x,y)	возвращает величину, абсолютное значение которой равно x, но знак которой соответствует знаку y
erf	функция ошибок
erfc	Дополнительная функция ошибок
exp2(x)	значение числа 2, возведённого в степень x, 2x
expm1(x)	значение функции ex − 1
fdim(x,y)	вычисление положительной разницы между x и y, fmax(x−y, 0)
fma(x,y,z)	значение функции (x * y) + z (см. FMA)
fmax(x,y)	наибольшее значение среди x и y
fmin(x,y)	наименьшее значение среди x и y
hypot(x,y)	гипотенуза, sqrt(x² + y²)
ilogb	экспонента числа с плавающей точкой, конвертированная в int
lgamma	натуральный логарифм абсолютного значения гамма-функции
llrint	округление до ближайшего целого (возвращает long long)
lrint	округление до ближайшего целого (возвращает long)
llround	округление до ближайшего целого в направлении от нуля (возвращает long long)
lround	округление до ближайшего целого в направлении от нуля (возвращает long)
log1p(x)	натуральный логарифм 1 + x
log2	логарифм по основанию 2
logb	целочисленная часть логарифма x по основанию 2
nan(s)	возвращает нечисловое значение 'Not a Number'
nearbyint	округление аргумента до целого значения в формате числа с плавающей точкой
nextafter(x,y)	следующий ближайшее представимое для x (по направлению к y)
nexttoward(x,y)	то же, что и nextafter, но y имеет тип long double
remainder(x,y)	вычисляет остаток от деления согласно стандарту IEC 60559
remquo(x,y,p)	то же, что и remainder, но сохраняет коэффициент по указателю p (как int)
rint	округление до целого (возвращает int) с вызовом ошибку inexact, если результат отличается от аргумента.
round	округление до целого (возвращает int)
scalbln(x,n)	x * FLT_RADIXn (n is long)
scalbn(x,n)	x * FLT_RADIXn (n is int)
trunc	округление до ближайшего целого числа в направлении к нулю
scalb(x,y)	x * FLT_RADIXy (x и y типа double)






*/

typedef Unifloat* (*caller_Uni) (Unifloat*);
typedef Unifloat* (*caller_IntUni) (int, Unifloat*);

static double call_Uni(caller_Uni function, double x)
{
	Unifloat* ux; 
	Unifloat* ures;
	double res;
	//printf("Unifloat function called!\n");
	Initialize();
	ux = convertDouble_UF(x);
//printf("x: %f\n", x);
	ures = (*function)(ux);
//PRINT_U("ux", ux);
//PRINT_U("ures", ures);
	res = convertUnifloat_Double(ures);

//printf("res: %f\n", res);
	delete_UFs(ux, ures, NULL);
	Finalize();
	return res;
}

static double call_IntUni(caller_IntUni function, int n, double x)
{
	Unifloat* ux; 
	Unifloat* ures;
	double res;
	//printf("Unifloat function called!\n");
	Initialize();
	ux = convertDouble_UF(x);
	ures = (*function)(n, ux);
	res = convertUnifloat_Double(ures);


	delete_UFs(ux, ures, NULL);
	Finalize();
	return res;
}

//SIN
long double sinl(long double x)
{return call_Uni(&sin_UF, x);}

double sin(double x)
{return call_Uni(&sin_UF, x);}

float sinf(float x)
{return call_Uni(&sin_UF, x);}


//ASIN
long double asinl(long double x)
{return call_Uni(&asin_UF, x);}

double asin(double x)
{return call_Uni(&asin_UF, x);}

float asinf(float x)
{return call_Uni(&asin_UF, x);}

//COS
long double cosl(long double x)
{return call_Uni(&cos_UF, x);}

double cos(double x)
{return call_Uni(&cos_UF, x);}

float cosf(float x)
{return call_Uni(&cos_UF, x);}

//ACOS
long double acosl(long double x)
{return call_Uni(&acos_UF, x);}

double acos(double x)
{return call_Uni(&acos_UF, x);}

float acosf(float x)
{return call_Uni(&acos_UF, x);}


//TAN
long double tanl(long double x)
{return call_Uni(&tan_UF, x);}

double tan(double x)
{return call_Uni(&tan_UF, x);}

float tanf(float x)
{return call_Uni(&tan_UF, x);}

//ATAN
long double atanl(long double x)
{return call_Uni(&atan_UF, x);}

double atan(double x)
{return call_Uni(&atan_UF, x);}

float atanf(float x)
{return call_Uni(&atan_UF, x);}

//EXP
long double expl(long double x)
{return call_Uni(&exp_UF, x);}

double exp(double x)
{return call_Uni(&exp_UF, x);}

float expf(float x)
{return call_Uni(&exp_UF, x);}


//LOG
long double logl(long double x)
{return call_Uni(&log_UF, x);}

double log(double x)
{return call_Uni(&log_UF, x);}

float logf(float x)
{return call_Uni(&log_UF, x);}


//j0
long double j0l(long double x)
{return call_Uni(&j0_UF, x);}

double j0(double x)
{return call_Uni(&j0_UF, x);}

float j0f(float x)
{return call_Uni(&j0_UF, x);}

//j1
long double j1l(long double x)
{return call_Uni(&j1_UF, x);}

double j1(double x)
{return call_Uni(&j1_UF, x);}

float j1f(float x)
{return call_Uni(&j1_UF, x);}

//jn
long double jnl(int n, long double x)
{return call_IntUni(&jn_UF, n, x);}

double jn(int n, double x)
{return call_IntUni(&jn_UF, n, x);}

float jnf(int n, float x)
{return call_IntUni(&jn_UF, n, x);}

//y0
long double y0l(long double x)
{return call_Uni(&y0_UF, x);}

double y0(double x)
{return call_Uni(&y0_UF, x);}

float y0f(float x)
{return call_Uni(&y0_UF, x);}

//y1
long double y1l(long double x)
{return call_Uni(&y1_UF, x);}

double y1(double x)
{return call_Uni(&y1_UF, x);}

float y1f(float x)
{return call_Uni(&y1_UF, x);}

//yn
long double ynl(int n, long double x)
{return call_IntUni(&yn_UF, n, x);}

double yn(int n, double x)
{return call_IntUni(&yn_UF, n, x);}

float ynf(int n, float x)
{return call_IntUni(&yn_UF, n, x);}


//GAMMA
long double gammal(long double x)
{return call_Uni(&gamma_UF, x);}

double gamma(double x)
{return call_Uni(&gamma_UF, x);}

float gammaf(float x)
{return call_Uni(&gamma_UF, x);}

//TGAMMA
long double tgammal(long double x)
{return call_Uni(&tgamma_UF, x);}

double tgamma(double x)
{return call_Uni(&tgamma_UF, x);}

float tgammaf(float x)
{return call_Uni(&tgamma_UF, x);}

//LGAMMA
long double lgammal(long double x)
{return call_Uni(&lgamma_UF, x);}

double lgamma(double x)
{return call_Uni(&lgamma_UF, x);}

float lgammaf(float x)
{return call_Uni(&lgamma_UF, x);}
